// © 2023, CGVR (https://cgvr.informatik.uni-bremen.de/),
// Author: Andre Mühlenbrock (muehlenb@uni-bremen.de)

// Define the menu height in pixel:
#define MENU_HEIGHT 105

// Include the GUI:
#define IMGUI_DEFINE_MATH_OPERATORS
#include <imgui_internal.h>
#include <imgui.h>
#include <backends/imgui_impl_glfw.h>
#include <backends/imgui_impl_opengl3.h>
#include <backends/imgui_impl_opengl3_loader.h>

// Include functions to display Mat4f and Vec4f in the GUI:
#include <imgui_cg1_helpers.h>
#include <GLFW/glfw3.h>
#include <iostream>

// Incude function to export image:
#define STB_IMAGE_WRITE_IMPLEMENTATION
#include <stb_image_write.h>

// Include math classes from the CG1 course:
#include "src/math/Vec4f.h"

// Include the raytracer:
#include "src/raytracer/Raytracer.h"

// Load scene:
#include "src/scene/SceneLoader.h"

#include <sstream>

using namespace std;
using namespace std::chrono;

/**
 * Initializes the application and the GUI
 */
int main(int, char**)
{    
    // Setup window:
    glfwSetErrorCallback([](int error, const char* description) {
        fprintf(stderr, "Glfw Error %d: %s\n", error, description);
    });

    if (!glfwInit())
        return 1;

    // Decide GL+GLSL versions:
    const char* glsl_version = "#version 330 core";
    glfwWindowHint(GLFW_CONTEXT_VERSION_MAJOR, 3);
    glfwWindowHint(GLFW_CONTEXT_VERSION_MINOR, 3);
    glfwWindowHint(GLFW_OPENGL_PROFILE, GLFW_OPENGL_CORE_PROFILE);
    glfwWindowHint(GLFW_OPENGL_FORWARD_COMPAT, GL_TRUE);

    // Create window with graphics context:
    GLFWwindow* window = glfwCreateWindow(1000, 520, "Computergrafik 2: Accelerated Raytracing", NULL, NULL);
    if (window == NULL)
        return 1;

    glfwMakeContextCurrent(window);
    glfwSwapInterval(1);

    // Setup Dear ImGui context:
    IMGUI_CHECKVERSION();
    ImGui::CreateContext();
    ImGuiIO& io = ImGui::GetIO(); (void)io;
    io.ConfigFlags |= ImGuiConfigFlags_NavEnableKeyboard;     // Enable Keyboard Controls
    io.ConfigWindowsMoveFromTitleBarOnly = true;                // Don't move windows when clicking within them.

    // Setup Platform/Renderer backends:
    ImGui_ImplGlfw_InitForOpenGL(window, true);
    ImGui_ImplOpenGL3_Init(glsl_version);

    // Load Roboto Font:
    io.Fonts->AddFontFromFileTTF(CMAKE_SOURCE_DIR "/lib/imgui-1.88/misc/fonts/Roboto-Medium.ttf", 16.0f);

    // Setup flags for inner windows:
    int windowFlags = ImGuiWindowFlags_NoResize | ImGuiWindowFlags_NoMove | ImGuiWindowFlags_NoScrollbar | ImGuiWindowFlags_NoCollapse;

    int renderResolution = 2048;
    int previewResolution = 512;

    // The raytracer:
    Raytracer raytracer(renderResolution, previewResolution);

    // Generating a texture for the raytraced image so that it can be shown in the window:
    unsigned int raytracedImageTexture;
    glGenTextures(1, &raytracedImageTexture);
    glBindTexture(GL_TEXTURE_2D, raytracedImageTexture);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

    Scene scene = SceneLoader::loadScene(CMAKE_SOURCE_DIR "/example_scenes/a-sphere-and-planes.json");
    raytracer.setScene(scene);
    std::string currentSceneName = "a_sphere_and_planes";

    // Last mouse position:
    ImVec2 lastMousePosition;

    // Last ADS type
    AcceleratorType lastAds = AcceleratorType::OctreeType;
    AcceleratorType currAds;

    // Main loop which is executed every frame until the window is closed:
    while (!glfwWindowShouldClose(window))
    {
        // Processes all glfw events:
        glfwPollEvents();

        // Get the size of the window:
        int display_w, display_h;
        glfwGetFramebufferSize(window, &display_w, &display_h);

#if defined(__APPLE__)
        display_w /= 2.0;
        display_h /= 2.0;
#endif

        // Calculate the size of the inner window:
        int innerWindowWidth = display_w;
        int innerWindowHeight = display_h - MENU_HEIGHT;
        int lowerWindowLimit = min(innerWindowWidth, innerWindowHeight);

        // Create the GUI:
        {
            // Start the Dear ImGui frame
            ImGui_ImplOpenGL3_NewFrame();
            ImGui_ImplGlfw_NewFrame();
            ImGui::NewFrame();
        }

        // Window for Raytracing:
        {
            ImGui::SetNextWindowPos(ImVec2(float(0.f),0.f));
            ImGui::SetNextWindowSize(ImVec2(float(innerWindowWidth),float(innerWindowHeight)));
            ImGui::Begin("Accelerated Raytracing", nullptr, windowFlags);

            std::vector<Vec4f>& image = raytracer.getDisplayImage();

            if(image.size() > 0){
                int imageResolution = int(sqrt(image.size()));

                // Updates the raytracedImageTexture:
                glBindTexture(GL_TEXTURE_2D, raytracedImageTexture);
                glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, imageResolution, imageResolution, 0, GL_RGBA, GL_FLOAT, &image[0]);

                ImGui::SetCursorPosX((innerWindowWidth - lowerWindowLimit) * 0.5f);
                ImGui::SetCursorPosY((innerWindowHeight - lowerWindowLimit) * 0.5f);
                ImGui::Image((intptr_t)raytracedImageTexture, ImVec2(float(lowerWindowLimit), float(lowerWindowLimit)));
            }

            bool mouseDown = false;
            bool mouseScrolled = false;
            // Update camera on mouse dragging (If mouse is down, add the difference to mouseMovementX
            // and mouseMovementY):
            ImVec2 winSize = ImGui::GetWindowSize();
            ImVec2 winPos = ImGui::GetWindowPos();
            if(ImGui::IsMouseDown(ImGuiMouseButton_Left)&& ImGui::IsMouseHoveringRect(winPos, winPos + winSize)){
                // Calculate delta between current mouse postion and the mouse position in the
                // last frame:
                ImVec2 mouseDelta = ImGui::GetMousePos() - lastMousePosition;

                // Rotate the camera:
                raytracer.getCamera().rotateCamera(mouseDelta.x, mouseDelta.y);
                mouseDown = true;
            }

            // Update camera distance when scrolling the mouse:
            if(ImGui::GetIO().MouseWheel != 0){
                raytracer.getCamera().multiplyDistance(ImGui::GetIO().MouseWheel > 0 ? 0.9f : 1.1f);
                mouseScrolled = true;
            }

            raytracer.setPreviewMode(mouseDown || mouseScrolled);

            ImGui::End();
        }

        // Settings:
        {
            ImGui::SetNextWindowPos(ImVec2(0.f,float(innerWindowHeight)));
            ImGui::SetNextWindowSize(ImVec2(float(display_w),float(MENU_HEIGHT)));
            ImGui::Begin("Settings", nullptr, windowFlags | ImGuiWindowFlags_NoTitleBar);

            ImGui::ProgressBar(raytracer.getProgress(), ImVec2(-FLT_MIN, 4), "");

            // Acceleration Data Structure
            ImGui::Text("Acceleration Data Structure: ");
            const char* ads[] = { "None", "Uniform Grid", "AABB-Tree", "kD-Tree", "Sphere-Tree", "Octree" };
            static const char *current_item = ads[0];
            static bool changed_ads = false;
            ImGui::SameLine();
            ImGui::SetNextItemWidth(150);
            if (ImGui::BeginCombo("##ads", current_item)) {
                changed_ads = true;
                for (int n = 0; n < IM_ARRAYSIZE(ads); n++) {
                    bool is_selected = (current_item == ads[n]);
                    if (ImGui::Selectable(ads[n], is_selected))
                        current_item = ads[n];
                    if (is_selected)
                        ImGui::SetItemDefaultFocus();
                }
                ImGui::EndCombo();
            } else if (changed_ads) {
                changed_ads = false;
                if (std::string(current_item).compare("None") == 0) {
                    currAds = AcceleratorType::NoneType;
                } else if (std::string(current_item).compare("Uniform Grid") == 0) {
                    currAds = AcceleratorType::UniformGridType;
                } else if (std::string(current_item).compare("AABB-Tree") == 0) {
                    currAds = AcceleratorType::AabbTreeType;
                } else if (std::string(current_item).compare("kD-Tree") == 0) {
                    currAds = AcceleratorType::KdTreeType;
                } else if (std::string(current_item).compare("Sphere-Tree") == 0) {
                    currAds = AcceleratorType::SphereTreeType;
                } else if (std::string(current_item).compare("Octree") == 0) {
                    currAds = AcceleratorType::OctreeType;
				}
                if (currAds != lastAds)
                    raytracer.setAccelerator( currAds );
                lastAds = currAds;
            }

            // Running Time
            ImGui::SameLine();
            auto run_time = raytracer.getRunningTime();
            std::stringstream stream;
            stream << std::fixed << std::setprecision(2) << run_time;
            ImGui::SetCursorPosX(float(display_w - 140));
            ImGui::Text("Running Time: %s s", stream.str().c_str());

            // Scene Loading
            ImGui::Text("Load Scene: ");

            bool disabledButtons = raytracer.getIsRendering();

            if (disabledButtons){
                ImGui::PushItemFlag(ImGuiItemFlags_Disabled, true);
                ImGui::PushStyleVar(ImGuiStyleVar_Alpha, ImGui::GetStyle().Alpha * 0.5f);
            }

            ImGui::SameLine();
            if(ImGui::Button("A Sphere and Planes")){
                Scene scene = SceneLoader::loadScene(CMAKE_SOURCE_DIR "/example_scenes/a-sphere-and-planes.json");
                previewResolution = 512; renderResolution = 2048;
                raytracer.setScene(scene);
                currentSceneName = "a_sphere_and_planes";
            }
            ImGui::SameLine();
            if(ImGui::Button("Metal Spheres")){
                Scene scene = SceneLoader::loadScene(CMAKE_SOURCE_DIR "/example_scenes/metal-spheres.json");
                previewResolution = 512; renderResolution = 2048;
                raytracer.setScene(scene);
                currentSceneName = "metal_spheres";
            }
            ImGui::SameLine();
            if(ImGui::Button("Glass Spheres")){
                Scene scene = SceneLoader::loadScene(CMAKE_SOURCE_DIR "/example_scenes/glass-spheres.json");
                previewResolution = 512; renderResolution = 2048;
                raytracer.setScene(scene);
                currentSceneName = "glass_spheres";
            }
            ImGui::SameLine();
            if(ImGui::Button("Mesh: Sphere")){
                Scene scene = SceneLoader::loadScene(CMAKE_SOURCE_DIR "/example_scenes/sphere.json");
                previewResolution = 128; renderResolution = 512;
                raytracer.setScene(scene);
                currentSceneName = "sphere";
            }
            ImGui::SameLine();
            if(ImGui::Button("Mesh: Cube")){
                Scene scene = SceneLoader::loadScene(CMAKE_SOURCE_DIR "/example_scenes/cube.json");
                previewResolution = 256; renderResolution = 1024;
                raytracer.setScene(scene);
                currentSceneName = "cube";
            }
            ImGui::SameLine();
            if(ImGui::Button("Mesh: Monkies")){
                Scene scene = SceneLoader::loadScene(CMAKE_SOURCE_DIR "/example_scenes/monkies.json");
                previewResolution = 64; renderResolution = 512;
                raytracer.setScene(scene);
                currentSceneName = "monkies";
            }
            ImGui::SameLine();
            if(ImGui::Button("Mesh: Many Spheres")){
                Scene scene = SceneLoader::loadScene(CMAKE_SOURCE_DIR "/example_scenes/many-spheres.json");
                previewResolution = 64; renderResolution = 512;
                raytracer.setScene(scene);
                currentSceneName = "monkies";
            }

            if (disabledButtons)
            {
                ImGui::PopItemFlag();
                ImGui::PopStyleVar();

                ImGui::SameLine();
                ImGui::SetCursorPosX(float(display_w - 60));
                if(ImGui::Button("Stop", ImVec2(50, 0))){
                    raytracer.interrupt();
                }
            } else {
                ImGui::SameLine();
                ImGui::SetCursorPosX(float(display_w - 90));
                if(ImGui::Button("Save Image", ImVec2(80, 0))){
                    std::vector<Vec4f>& image = raytracer.getDisplayImage();
                    std::vector<unsigned char> conv_image(image.size() * 4);

                    #pragma omp parallel num_threads(16)
                    {
                        int size = int(image.size());
                        #pragma omp for
                        for(int i = 0; i < size; ++i){
                            Vec4f v = image[i];
                            conv_image[i*4] = (unsigned char)(min(255.f, v.x * 255));
                            conv_image[i*4+1] = (unsigned char)(min(255.f, v.y * 255));
                            conv_image[i*4+2] = (unsigned char)(min(255.f, v.z * 255));
                            conv_image[i*4+3] = 255;
                        }
                    }

                    std::string sourceFolder = CMAKE_SOURCE_DIR;
                    std::string filename = sourceFolder + "/rendered_images/" + currentSceneName + "_" + std::to_string(std::time(0)) + ".png";
                    stbi_write_png(filename.c_str(), renderResolution, renderResolution, 4, &conv_image[0], renderResolution * 4);
                }
            }

            ImGui::Separator();

            ImGui::Text("Render Resolution:");
            ImGui::SameLine();
            ImGui::SetNextItemWidth(100);
            ImGui::InputInt("##1", &renderResolution);
            ImGui::SameLine();
            ImGui::Text(" | Preview Resolution:");
            ImGui::SameLine();
            ImGui::SetNextItemWidth(100);
            ImGui::InputInt("##2", &previewResolution);


            ImGui::End();
        }

        // Render the GUI and draw it to the screen:
        ImGui::Render();
        ImGui_ImplOpenGL3_RenderDrawData(ImGui::GetDrawData());

        // Swap Buffers:
        glfwSwapBuffers(window);

        lastMousePosition = ImGui::GetMousePos();
    }

    // Cleanup
    ImGui_ImplOpenGL3_Shutdown();
    ImGui_ImplGlfw_Shutdown();
    ImGui::DestroyContext();

    glfwDestroyWindow(window);
    glfwTerminate();

    return 0;
}

