// © 2023, CGVR (https://cgvr.informatik.uni-bremen.de/),
// Author: Andre Mühlenbrock (muehlenb@uni-bremen.de)

#pragma once

#include <vector>
#include <thread>

#include <Semaphore.h>

#include "Ray.h"

#include "src/math/Vec4f.h"

#include "src/scene/Scene.h"

/**
 * A raytracer which renders a scene.
 */
class Raytracer
{
    /** The scene defining a camera, surface and lights */
    Scene scene;

    /** Background color when no object was hit by a ray */
    Vec4f backgroundColor = Vec4f(0.5, 0.5, 0.5, 1.0f);

    /** The preview image consisting of RGBA-pixels (one pixel is a Vec4f where xyzw is rgba) */
    std::vector<Vec4f> previewImage;

    /** The rendered image consisting of RGBA-pixels (one pixel is a Vec4f where xyzw is rgba) */
    std::vector<Vec4f> renderedImage;

    /** Render resolution */
    int& renderResolution;

    /** Preview resolution */
    int& previewResolution;

    /** Max ray depth */
    unsigned int maxRayDepth = 10;

    /** We use a separate thread for raytracing so the resulting image is updated in realtime */
    std::thread raytraceThread;

    /** Raytrace semaphore */
    Semaphore raytraceSemaphore;

    /** Should the preview image currently be displayed? */
    bool displayPreviewImage = false;

    /** Random pixel ID map which is used to make the order of rendering random */
    std::vector<unsigned int> previewPixelIDMap;

    /** Random pixel ID map which is used to make the order of rendering random */
    std::vector<unsigned int> renderPixelIDMap;

    /** Raytracer waits for termination */
    bool waitForTerminate = false;

    /** Preview mode is activated when camera is rotated */
    bool previewMode = false;

    /** Is currently rendering in progress? */
    bool isRendering = false;

    /** Render progress */
    float progress = 0.f;

    /** Render time */
    float runningTime = 0.f;

    /** Should stop the active raytracing in next iteration? */
    bool shouldInterrupt = false;

    std::mutex renderLock;

    AcceleratorType accelerator;

    /**
     * Renders the image by generating a primary ray for each pixel using the camera and
     * tracing along them using the traceRay and shade function.
     */
    void render(bool renderInPreviewMode);

    void initPixelIDMap(std::vector<unsigned int>& pixelIDMap, unsigned int size);
public:
    /** Constructor */
    Raytracer(int& renderResolution, int& previewResolution);

    /** Destructor */
    ~Raytracer();

    /**
     * Returns the image that should be shown.
     */
    std::vector<Vec4f>& getDisplayImage(){ return displayPreviewImage ? previewImage : renderedImage; }

    /**
     * Set whether the scene should be rendered in preview mode or not.
     */
    void setPreviewMode(bool mode);

    /**
     * Set the scene.
     */
    void setScene(Scene scene);

    /**
     * Returns whether currently is something rendered.
     */
    bool getIsRendering(){ return isRendering; };

    /**
     * Returns the progress of full rendering in a range from 0 to 1.
     */
    float getProgress();

    /** Return the time since rendering of this frame started (in ms). */
    float getRunningTime();

    /**
     * Returns the camera.
     */
    Camera& getCamera(){ return scene.camera; }

    /**
     * Interrupts the active raytracing
     */
    void interrupt(){ shouldInterrupt = true; }

    /**
     * Copy preview to full.
     */
    void copyPreviewToFull();

    /**
     * Initializes an image to render on with the given width and height.
     */
    void resizeImagesIfNecessary();

    /**
     * Traces a given primary ray against the given surace definition and returns a color.
     * Depending on the depth, also secondary rays and shadow feelers are generated and
     * traced too.
     *
     * This function is called once for every pixel of the final rendered image.
     *
     * @param ray       The primary ray.
     * @param depth     Recursion depth for the rays.
     *                  1: no secondary rays,
     *                  2: one level refraction/reflection rays
     *                  3: two level of refraction/reflection rays
     *                  ...
     *
     * @return The color of the pixel.
     */
    Vec4f traceRay(Ray ray, unsigned int depth);

    /**
     * Shades an intersection point of a primary or secondary ray. This means the color
     * of outgoing light is calculated by the incoming light and the normal of the surface
     * with the intersection point.
     *
     * @param intersection      Intersection of the hitpoint.
     * @param ray               Ray that intersected the surface.
     *
     * @return Color of the intersection point.
     */
    Vec4f shade(Intersection& intersection, Ray& ray);

    void setAccelerator(AcceleratorType acc) { accelerator = acc; };
};
