#include "Mesh.h"
#include "Triangle.h"

#include <string>
#include <sstream>

#define TINYOBJLOADER_IMPLEMENTATION
#include "tiny_obj_loader.h"

Mesh::Mesh(Vec4f pTranslation, Vec4f pScale, Shader pShader, std::string fileName)
{
    position = pTranslation;
    scale = pScale;
    M_WfL = Mat4f::translation(pTranslation.x,pTranslation.y,pTranslation.z) * Mat4f::scale(pScale.x,pScale.y,pScale.z);
    shader = pShader;
    meshFile = fileName;

    initializeTriangles();
}

void Mesh::initializeTriangles()
{
    tinyobj::attrib_t attrib;
    std::vector<tinyobj::shape_t> shapes;
    std::vector<tinyobj::material_t> materials;
    std::string warn, err;
    std::string filePath = CMAKE_SOURCE_DIR + std::string("/") + meshFile;
    bool retval = tinyobj::LoadObj( &attrib, &shapes, &materials, &warn, &err, filePath.c_str() );
    if (!warn.empty())
        std::cout << "[Mesh] Warning in LoadObj: " << warn << std::endl;
    if (!err.empty())
        std::cout << "[Mesh] Error in LoadObj: " << err << std::endl;
    if (!retval)
        std::cout << "[Mesh] LoadObj failed!" << std::endl;

    for (int i=0; i+2<attrib.vertices.size(); i+=3)
        vertices.push_back( Vec4f(attrib.vertices[i+0],
                                  attrib.vertices[i+1],
                                  attrib.vertices[i+2],
                                  1 ) );
    for (int i=0; i+2<attrib.normals.size(); i+=3)
        normals.push_back( Vec4f(attrib.normals[i+0],
                                 attrib.normals[i+1],
                                 attrib.normals[i+2],
                                 0 ).normalized() );
    for (int i=0; i<shapes.size(); ++i) {
        for (int j=0; j+2<shapes[i].mesh.indices.size(); j+=3) {
            tinyobj::index_t i0 = shapes[i].mesh.indices[j+0];
            tinyobj::index_t i1 = shapes[i].mesh.indices[j+1];
            tinyobj::index_t i2 = shapes[i].mesh.indices[j+2];
            auto triangle = std::make_shared<Triangle>(
                vertices[i0.vertex_index],
                vertices[i1.vertex_index],
                vertices[i2.vertex_index],
                normals[i0.normal_index],
                normals[i1.normal_index],
                normals[i2.normal_index],
                shader,
                M_WfL
            );
            triangles.push_back( triangle );
        }
    }
}

std::vector<Vec4f> Mesh::getVertices()
{
    return vertices;
}

std::vector<Vec4f> Mesh::getNormals()
{
    return normals;
}

std::vector<std::shared_ptr<Surface>> Mesh::getTriangles()
{
    return triangles;
}

Mat4f Mesh::getWorldFromLocal()
{
    return M_WfL;
}
