#pragma once

#include "src/scene/Surface.h"
#include "src/shader/Shader.h"
#include "src/math/Mat4f.h"

#include <memory>

class Mesh : std::enable_shared_from_this<Mesh> {
    /** World from local transformation. */
    Mat4f M_WfL;

    /** World from local translation. */
    Vec4f position;

    /** World from local scaling. */
    Vec4f scale;

    /** Shader that is used to shade a pixel */
    Shader shader;

    /** Array of all mesh vertices. Indexed by Triangle::vertexIndexX. */
    std::vector<Vec4f> vertices;

    /** Array of all mesh vertex normals. Indexed by Triangle::normalIndexX. */
    std::vector<Vec4f> normals;

    /** Array of all triangle surface objects. Indexed by Triangle::faceIndex. */
    std::vector<std::shared_ptr<Surface>> triangles;

    /** File path to the mesh file. */
    std::string meshFile;

    /** Read the mesh file and create triangle objects */
    void initializeTriangles();

public:
    /**
     * Creates a mesh and sets transformation based on the provided parameters.
     *
     * @param pTranslation Translation vector to transform from local to world space.
     * @param pScale Scaling vector to transform from local to world space.
     * @param pShader Shader.
     * @param fileName File name of the mesh file to be loaded.
     */
    Mesh(Vec4f pTranslation, Vec4f pScale, Shader pShader, std::string fileName);

    // Getter for the list of all mesh vertices
    std::vector<Vec4f> getVertices();

    // Getter for the list of all mesh normals
    std::vector<Vec4f> getNormals();

    // Getter for the list of all mesh triangles
    std::vector<std::shared_ptr<Surface>> getTriangles();

    // Getter for world from local transformation
    Mat4f getWorldFromLocal();
};