#include "SurfaceList.h"

#include "UniformGrid.h"
#include "AabbTree.h"
#include "SphereTree.h"
#include "KdTree.h"
#include "Octree.h"

#include <chrono>
#include <format>

void SurfaceList::addSurface(std::shared_ptr<Surface> surface){
    surfaces.push_back(surface);
    aabb.extendBy(surface->getAabb());
}

void SurfaceList::addSurfaces(std::vector<std::shared_ptr<Surface>> pSurfaces){
    for (auto surface : pSurfaces)
        addSurface(surface);
}

Intersection SurfaceList::intersect(Ray ray, float t0, float t1){
    Intersection result;

    if (acceleratorType != AcceleratorType::NoneType) {
        result = accelerator->intersect( ray, t0, t1 );
    } else {
        for(std::shared_ptr<Surface>& surface : surfaces) {
            result.combineClosest( surface->intersect(ray, t0, t1) );
            // if (result.hasIntersection)
            //     t1 = result.t;
        }
    }

    return result;
}

void SurfaceList::setAccelerator( AcceleratorType acc )
{
    if (acc < AcceleratorType::NoneType || AcceleratorType::OctreeType < acc)
        return;
    if (acceleratorType == acc && accelerator)
        return;

    // auto const time = std::chrono::current_zone()->to_local(std::chrono::system_clock::now());
    // std::cout <<  "["<<std::format("{:%T}", time)<<"] setAccelerator("<<acc<<")" << std::endl;
    acceleratorType = acc;
    initializeAccelerator( acc );
}

void SurfaceList::initializeAccelerator( AcceleratorType acc )
{
    destroyAccelerator();
    acceleratorType = acc;
    auto start_time = std::chrono::high_resolution_clock::now();

    if (acceleratorType == AcceleratorType::UniformGridType)
		accelerator = new UniformGrid( this );
    else if (acceleratorType == AcceleratorType::AabbTreeType)
		accelerator = new AabbTree( this );
    else if (acceleratorType == AcceleratorType::SphereTreeType)
		accelerator = new SphereTree( this );
    else if (acceleratorType == AcceleratorType::KdTreeType)
		accelerator = new KdTree( this );
    else if (acceleratorType == AcceleratorType::OctreeType)
		accelerator = new Octree( this );

    auto end_time = std::chrono::high_resolution_clock::now();
    std::cout << "ADS Generation took " << std::chrono::duration_cast<std::chrono::milliseconds>(end_time-start_time).count()/1000.0 << " seconds." << std::endl;
}

void SurfaceList::destroyAccelerator()
{
    acceleratorType = NoneType;
    delete accelerator;
    accelerator = nullptr;
}

Aabb SurfaceList::getAabb() const
{
    return aabb;
}

std::vector<std::shared_ptr<Surface>> SurfaceList::getSurfaces() const
{
    return surfaces;
}