#include "Triangle.h"

#include "src/shader/Shader.h"
#include "src/scene/Mesh.h"

Triangle::Triangle(Vec4f v0, Vec4f v1, Vec4f v2, Vec4f n0, Vec4f n1, Vec4f n2, Shader pShader, Mat4f M)
{
    vs[0] = M*v0;
    vs[1] = M*v1;
    vs[2] = M*v2;
    calcAabb();
    ns[0] = M*n0;
    ns[1] = M*n1;
    ns[2] = M*n2;
    shader = pShader;
}

Intersection Triangle::intersect(Ray ray, float tStart, float tEnd)
{
    Intersection intersection;

    auto u = vs[1] - vs[0];
    auto v = vs[2] - vs[0];
    auto n = u.cross(v).normalized();

    auto pvec = ray.direction.cross(v);

    auto det = u.dot(pvec);

#define EPSILON 0.00001
    if (-EPSILON < det && det < EPSILON)
        return intersection;
    auto inv_det = 1.0 / det;

    auto tvec = ray.origin - vs[0];
    auto tu = tvec.dot(pvec) * inv_det;
    if (tu < 0 || tu > 1)
        return intersection;

    auto qvec = tvec.cross(u);
    auto tv = ray.direction.dot(qvec) * inv_det;
    if (tv < 0 || tu + tv > 1)
        return intersection;

    // calculate t, ray intersects triangle's plane
    auto t = v.dot(qvec) * inv_det;
    auto point_in_plane = ray.evaluate(t);

    if (t < tStart || t > tEnd)
        return intersection;

    // normal interpoaltion with barycentric coordinates
    float uu = u.dot(u);
    float uv = u.dot(v);
    float vv = v.dot(v);

    Vec4f w = point_in_plane - vs[0];
    float wu = w.dot(u);
    float wv = w.dot(v);
    float D = uv*uv - uu*vv;

    float beta = (uv*wv - vv*wu) / D;
    float gamma = (uv*wu - uu*wv) / D;

    intersection.hasIntersection = true;
    intersection.t = t;
    intersection.hitPoint = point_in_plane;
    intersection.normal = ((1-beta-gamma)*ns[0] + beta*ns[1] + gamma*ns[2]).normalized();
    intersection.shader = &shader;

    return intersection;
}

void Triangle::calcAabb()
{
    aabb.extendBy(vs[0]);
    aabb.extendBy(vs[1]);
    aabb.extendBy(vs[2]);
}

Aabb Triangle::getAabb() const
{
    return aabb;
}