// © 2023, CGVR (https://cgvr.informatik.uni-bremen.de/),
// Author: Andre Mühlenbrock (muehlenb@uni-bremen.de)

#pragma once

#include "src/math/Vec4f.h"

#include "src/raytracer/Intersection.h"
#include "src/raytracer/Ray.h"
#include "src/scene/PointLight.h"

class Shader {
public:
    /** Ambient color */
    Vec4f ambientColor = Vec4f(0.0f, 0.0f, 0.0f);

    /** Diffuse color */
    Vec4f diffuseColor = Vec4f(0.6f, 0.6f, 0.6f);

    /** Shininess */
    float shininess = 200.f;

    /** Reflective color */
    Vec4f reflectiveColor = Vec4f(0.4f, 0.4f, 0.4f);

    /** Refractive color */
    Vec4f refractiveColor = Vec4f(0.0f, 0.0f, 0.0f);

    /** Refractive index */
    float refractiveIndex = 0.96f;

    /**
     * Shades the intersected surface using the given light source. This
     * function is called for every light once.
     *
     * @param light             Point light source which should be used to
     *                          shade the intersected surface.
     * @param intersection      Intersection point which is should be shaded.
     * @param ray               Ray which intersected the surface.
     *
     * @return The color.
     */
    Vec4f shade(PointLight& light, Intersection& intersection, Ray& ray);

    /**
     * Shade ambient.
     */

    Vec4f shadeAmbient(Vec4f color);

    /**
     * Returns the color of this shader/material.
     */
    Vec4f getAmbientColor();

    /**
     * Returns the reflected color (multiplier) of this shader.
     *
     * This is the vector the color of secondary rays is multiplied with.
     */
    Vec4f getReflectiveColor();

    /**
     * Returns the refractive color (multiplier) of this shader.
     *
     * This is the vector the color of secondary rays is multiplied with.
     */
    Vec4f getRefractiveColor();

    /**
     * Returns the refractive index of this shader/material.
     */
    float getRefractiveIndex();

    /**
     * Returns whether this shader has a reflective component (and whether we
     * have to shoot secondary rays for this).
     */
    bool hasReflective();

    /**
     * Returns whether this shader has a refractive component (and whether
     * we have to shoot secondary rays for this).
     */
    bool hasRefractive();
};
